<?php
/**
 * ClassyLlama_AvaTax
 *
 * NOTICE OF LICENSE
 *
 * This source file is subject to the Open Software License (OSL 3.0)
 * that is bundled with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://opensource.org/licenses/osl-3.0.php
 *
 * @copyright  Copyright (c) 2016 Avalara, Inc.
 * @license    http://opensource.org/licenses/osl-3.0.php Open Software License (OSL 3.0)
 */

namespace ClassyLlama\AvaTax\Framework\Interaction\Cacheable;

use AvaTax\GetTaxRequest;
use AvaTax\GetTaxResult;
use ClassyLlama\AvaTax\Framework\Interaction\MetaData\MetaDataObject;
use ClassyLlama\AvaTax\Framework\Interaction\MetaData\MetaDataObjectFactory;
use ClassyLlama\AvaTax\Framework\Interaction\Tax;
use ClassyLlama\AvaTax\Helper\Config;
use ClassyLlama\AvaTax\Model\Logger\AvaTaxLogger;
use Magento\Framework\App\CacheInterface;
use Magento\Framework\Exception\LocalizedException;
use Magento\Framework\Phrase;
use Magento\Framework\Serialize\Serializer\Serialize;

class TaxService
{
    /**
     * 1 day in seconds
     */
    const CACHE_LIFETIME = 86400;

    /**
     * @var CacheInterface
     */
    protected $cache = null;

    /**
     * @var AvaTaxLogger
     */
    protected $avaTaxLogger = null;

    /**
     * @var Tax
     */
    protected $taxInteraction = null;

    /**
     * @var MetaDataObject
     */
    protected $metaDataObject = null;

    /**
     * @var null
     */
    protected $type = null;

    /**
     * @var Serialize
     */
    protected $serializer;

    /**
     * @param CacheInterface $cache
     * @param AvaTaxLogger $avaTaxLogger
     * @param Tax $taxInteraction
     * @param MetaDataObjectFactory $metaDataObjectFactory
     * @param null $type
     * @param Serialize $serializer
     */
    public function __construct(
        CacheInterface $cache,
        AvaTaxLogger $avaTaxLogger,
        Tax $taxInteraction,
        Serialize $serializer,
        MetaDataObjectFactory $metaDataObjectFactory,
        $type = null
    ) {
        $this->cache = $cache;
        $this->avaTaxLogger = $avaTaxLogger;
        $this->taxInteraction = $taxInteraction;
        $this->serializer = $serializer;
        $this->metaDataObject = $metaDataObjectFactory->create(
            ['metaDataProperties' => \ClassyLlama\AvaTax\Framework\Interaction\Tax::$validFields]
        );
    }

    /**
     * Cache validated response
     *
     * @param GetTaxRequest $getTaxRequest
     * @param $storeId
     * @param bool $useCache
     * @return GetTaxResult
     * @throws LocalizedException
     */
    public function getTax(GetTaxRequest $getTaxRequest, $storeId, $useCache = false)
    {
        $cacheKey = $this->getCacheKey($getTaxRequest) . $storeId;
        $cacheData = $this->cache->load($cacheKey);
        if (false === $cacheData || null === $cacheData || '' === $cacheData) {
            $getTaxResult='';
        } else {
            $getTaxResult = $this->serializer->unserialize($cacheData);
        }

        if ($getTaxResult instanceof GetTaxResult && $useCache) {
            $this->avaTaxLogger->addDebug('Loaded \AvaTax\GetTaxResult from cache.', [
                'result' => var_export($getTaxResult, true),
                'cache_key' => $cacheKey
            ]);
            return $getTaxResult;
        }

        $getTaxResult = $this->taxInteraction->getTaxService($this->type, $storeId)->getTax($getTaxRequest);
        $this->avaTaxLogger->addDebug('Loaded \AvaTax\GetTaxResult from SOAP.', [
            'request' => var_export($getTaxRequest, true),
            'result' => var_export($getTaxResult, true),
        ]);

        // Only cache successful requests
        if ($useCache && $getTaxResult->getResultCode() == \AvaTax\SeverityLevel::$Success) {
            $serializedGetTaxResult = $this->serializer->serialize($getTaxResult);
            $this->cache->save(
                $serializedGetTaxResult,
                $cacheKey,
                [Config::AVATAX_CACHE_TAG],
                self::CACHE_LIFETIME
            );
        }
        return $getTaxResult;
    }

    /**
     * Create cache key by calling specified methods and concatenating and hashing
     *
     * @param $object
     * @return string
     * @throws LocalizedException
     */
    protected function getCacheKey($object)
    {
        return $this->metaDataObject->getCacheKeyFromObject($object);
    }

    /**
     * Pass all undefined method calls through to Tax Service
     *
     * @param $name
     * @param array $arguments
     * @return mixed
     */
    public function __call($name , array $arguments)
    {
        return call_user_func_array([$this->taxInteraction->getTaxService($this->type), $name], $arguments);
    }

}
